------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.7     071213
--
--           'K e r n e l . T a s k _ S e t s . O p e r a t i o n s'
--
--                                 Body
--
--  File 'kernel-task_sets-operations.adb'                          By MAR.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
-------------------------------------------------------------------------------
with Singly_Linked_Lists.Resources;
pragma Elaborate_All (Singly_Linked_Lists.Resources);
with Kernel.Timers.Internals;
with Kernel.Debug;
with Kernel.Group_Clocks.Internals;

package body Kernel.Task_Sets.Operations is

   package TS       renames Kernel.Task_Sets;
   package GCLK     renames Kernel.Group_Clocks;
   package GCLKI    renames Kernel.Group_Clocks.Internals;
   package TMRI     renames Kernel.Timers.Internals;
   package TC_Lists renames Kernel.Task_Containers_Lists;
   package CP       renames Configuration_Parameters;
   package DBG      renames Kernel.Debug;

   Initialized : Boolean := False;

   use type Kernel.Task_Container_Ac, Kernel.Task_Id, Kernel.Task_Set_Id,
       Kernel.CPU_Time_Timed_Event_Ac;

   ----------------------------------------------
   --  Pools of Task_Sets and Task_Containers  --
   ----------------------------------------------
   package Task_Sets_Pool is
     new Kernel.Task_Sets_Lists.Resources (Kernel.Task_Sets_Lists.Element);
   Free_Task_Sets : array (0 .. CP.Num_Task_Sets_Mx - 1) of
     aliased Kernel.Task_Sets_Lists.Element;

   package Task_Containers_Pool is
     new TC_Lists.Resources (Kernel.Task_Containers_Lists.Element);
   Free_Task_Containers : array (0 .. CP.Num_User_Tasks_Mx - 1) of
     aliased Kernel.Task_Containers_Lists.Element;

   --------------------------------
   --  Is_In_The_Group_Of_Tasks  --
   --------------------------------
   --  Internal function to be used for debug purposes
   function Is_In_The_Group_Of_Tasks (L : TC_Lists.List;
                                      T : Kernel.Task_Id) return Boolean;
   function Is_In_The_Group_Of_Tasks (L : TC_Lists.List;
                                      T : Kernel.Task_Id) return Boolean is
      TC : Kernel.Task_Container_Ac;
   begin
      if not Configuration_Parameters.Use_Task_Sets'First then
         return False;
      end if;

      TC := TC_Lists.Head (L);
      while TC /= null loop
         if TC.T = T then
            return True;  --  Task found in list
         end if;
         TC := TC_Lists.Next (TC);
      end loop;

      return False;  --  Task NOT found in list
   end Is_In_The_Group_Of_Tasks;

   -------------------------------
   --  Marte_Thread_Set_Create  --
   -------------------------------
   function Marte_Thread_Set_Create
     (Set : access Kernel.Task_Set_Id)
      return Int is
      Flags : Integer;
   begin
      pragma Debug (DBG.Assert (Initialized));

      if not Configuration_Parameters.Use_Task_Sets'First then
         return OPERATION_NOT_SUPPORTED;
      end if;

      Kernel.Enter_Critic_Section (Flags);

      Set.all := Task_Sets_Pool.Request;

      if Set.all = null then
         --  No resources free
         Kernel.Leave_Critic_Section (Flags);
         return RESOURCE_TEMPORARILY_UNAVAILABLE;
      end if;

      Task_Set_Base (Set.all.all) := (Set           => TC_Lists.Null_List,
                                      Consumed_Time => 0,
                                      Group_TE_Ac   => null,
                                      Iterator      => null,
                                      Magic         => Kernel.TS_INITIALIZED);

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Marte_Thread_Set_Create;

   --------------------------------
   --  Marte_Thread_Set_Destroy  --
   --------------------------------
   function Marte_Thread_Set_Destroy
     (Set : Kernel.Task_Set_Id)
      return Int is
      Flags : Integer;
      Ret : Int;
      use type Int;
   begin
      pragma Debug (DBG.Assert (Initialized));

      if not Configuration_Parameters.Use_Task_Sets'First then
         return OPERATION_NOT_SUPPORTED;
      end if;

      Kernel.Enter_Critic_Section (Flags);

      --  Empty task group
      Ret := Marte_Thread_Set_Empty (Set);
      pragma Debug (DBG.Assert (Ret = 0));

      --  Delete task group
      Set.all.Magic := Kernel.TS_NOT_INITIALIZED;
      Task_Sets_Pool.Release (Set);

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Marte_Thread_Set_Destroy;

   ------------------------------
   --  Marte_Thread_Set_Empty  --
   ------------------------------
   function Marte_Thread_Set_Empty
     (Set : Kernel.Task_Set_Id)
      return Int is
      Flags : Integer;
      TC : Kernel.Task_Container_Ac;
   begin
      pragma Debug (DBG.Assert (Initialized));

      if not Configuration_Parameters.Use_Task_Sets'First then
         return OPERATION_NOT_SUPPORTED;
      end if;

      Kernel.Enter_Critic_Section (Flags);
      if not TS.Task_Set_OK (Kernel.Task_Set_Base_Ac (Set)) then
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;

      pragma Debug (DBG.Assert (Set.Group_TE_Ac = null or else
                                GCLK.Group_TE_OK (Set.Group_TE_Ac)));

      --  Release TCs
      TC := TC_Lists.Head (Set.Set);
      while TC /= null loop
         pragma Debug (DBG.Assert (Kernel.Task_Set_Base_Ac (Set) = TC.T.Set));
         TC_Lists.Dequeue_Head (Set.Set);
         if Set.Group_TE_Ac /= null then
            GCLKI.Del_Task (TC.T, Set.Group_TE_Ac);
            TC.T.Set := null;
         end if;
         Task_Containers_Pool.Release (TC);
         TC := TC_Lists.Head (Set.Set);
      end loop;

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Marte_Thread_Set_Empty;

   -----------------------------------
   --  Marte_Thread_Set_Add_Thread  --
   -----------------------------------
   function Marte_Thread_Set_Add_Thread
     (Set : Kernel.Task_Set_Id;
      T   : Kernel.Task_Id)
      return Int is
      Flags : Integer;
      TC : Kernel.Task_Container_Ac;
   begin
      pragma Debug (DBG.Assert (Initialized));

      if not Configuration_Parameters.Use_Task_Sets'First then
         return OPERATION_NOT_SUPPORTED;
      end if;

      Kernel.Enter_Critic_Section (Flags);
      --  Check errors
      if not TS.Task_Set_OK (Kernel.Task_Set_Base_Ac (Set))
        or else not Kernel.Task_OK (T)
        or else (T.Set /= null and Kernel.Task_Set_Base_Ac (Set) /= T.Set)
      then
         --  Invalid argument
         --   * invalid Set
         --   * invalid T
         --   * Task already in a set (different from 'Set'). This is a
         --   restrictive limitation that could be removed in teh future
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;

      if Kernel.Task_Set_Base_Ac (Set) /= T.Set then
         pragma Debug (DBG.Assert (not Is_In_The_Group_Of_Tasks (Set.Set, T)));

         --  Request "Task_Container"
         TC := Task_Containers_Pool.Request;
         if TC = null then
            Kernel.Leave_Critic_Section (Flags);
            return RESOURCE_TEMPORARILY_UNAVAILABLE;
         end if;

         --  Add T to Set
         TC.T := T;
         TC_Lists.Enqueue_Head (TC, Set.Set);

         --  Link Set with T
         T.Set := Kernel.Task_Set_Base_Ac (Set);

      else
         --  If T is already in Set nothing has to be done
         pragma Debug (DBG.Assert (Is_In_The_Group_Of_Tasks (Set.Set, T)));
         null;
      end if;

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Marte_Thread_Set_Add_Thread;

   -----------------------------------
   --  Marte_Thread_Set_Del_Thread  --
   -----------------------------------
   function Marte_Thread_Set_Del_Thread
     (Set : Kernel.Task_Set_Id;
      T   : Kernel.Task_Id)
      return Int is
      Flags : Integer;
      TC : Kernel.Task_Container_Ac;
   begin
      pragma Debug (DBG.Assert (Initialized));

      if not Configuration_Parameters.Use_Task_Sets'First then
         return OPERATION_NOT_SUPPORTED;
      end if;

      Kernel.Enter_Critic_Section (Flags);
      --  Check errors
      if not TS.Task_Set_OK (Kernel.Task_Set_Base_Ac (Set))
        or else not Kernel.Task_OK (T)
        or else Kernel.Task_Set_Base_Ac (Set) /= T.Set
      then
         --  Invalid argument
         --   * invalid Set
         --   * invalid T
         --   * T not in Set
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;

      --  Delete task from the set
      TC := TC_Lists.Head (Set.Set);
      loop
         if TC = null then
            --  This should NEVER happen: task is not in the list !!
            pragma Debug (DBG.Assert (False));
            Kernel.Leave_Critic_Section (Flags);
            return NO_SUCH_PROCESS;
         end if;

         if TC.T = T then
            --  Task found
            TC_Lists.Dequeue (TC, Set.Set);
            Task_Containers_Pool.Release (TC);
            exit;
         end if;

         TC := TC_Lists.Next (TC);
      end loop;
      pragma Debug (DBG.Assert (not Is_In_The_Group_Of_Tasks (Set.Set, T)));

      --  Set associated with groupclock?
      if Set.Group_TE_Ac /= null then
         pragma Debug (DBG.Assert (GCLK.Group_TE_OK (Set.Group_TE_Ac)));
         --  Unlink T from the Groupclock
         GCLKI.Del_Task (T, Set.Group_TE_Ac);
      end if;

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Marte_Thread_Set_Del_Thread;

   ------------------------------
   --  Marte_Thread_Set_First  --
   ------------------------------
   function Marte_Thread_Set_First (Set : Kernel.Task_Set_Id;
                                    T   : access Kernel.Task_Id) return Int is
   begin
      pragma Debug (DBG.Assert (Initialized));

      --  Not implemented yet
      pragma Debug (DBG.Assert (False));
      return 0;
   end Marte_Thread_Set_First;

   -----------------------------
   --  Marte_Thread_Set_Next  --
   -----------------------------
   function Marte_Thread_Set_Next (Set : Kernel.Task_Set_Id;
                                   T   : access Kernel.Task_Id) return Int is
   begin
      pragma Debug (DBG.Assert (Initialized));

      --  Not implemented yet
      pragma Debug (DBG.Assert (False));
      return 0;
   end Marte_Thread_Set_Next;

   -------------------------------
   --  Marte_Thread_Set_Getset  --
   -------------------------------
   function Marte_Thread_Set_Getset (Set : access Kernel.Task_Set_Id;
                                     T   : Kernel.Task_Id) return Int is
   begin
      pragma Debug (DBG.Assert (Initialized));

      --  Not implemented yet
      pragma Debug (DBG.Assert (False));
      return 0;
   end Marte_Thread_Set_Getset;

   -----------------------------
   --  Marte_Getgroupclockid  --
   -----------------------------
   function Marte_Getgroupclockid
     (Set   : Kernel.Task_Set_Id;
      Clock : access Kernel.Clock_Id)
      return Int is
      Flags : Integer;
   begin
      pragma Debug (DBG.Assert (Initialized));

      if not Configuration_Parameters.Use_Task_Sets'First
        or not Configuration_Parameters.Use_CPU_Time_Clocks_And_Timers
        or not Configuration_Parameters.Use_Group_Clocks'First
      then
         return OPERATION_NOT_SUPPORTED;
      end if;

      Kernel.Enter_Critic_Section (Flags);
      --  Check errors
      if not TS.Task_Set_OK (Kernel.Task_Set_Base_Ac (Set))
      then
         --  Invalid argument
         --   * invalid Set
         --   * null clock pointer
         Kernel.Leave_Critic_Section (Flags);
         return INVALID_ARGUMENT;
      end if;

      --  Return Clock
      Clock.all := TMRI.To_Clock_Id (Set);

      Kernel.Leave_Critic_Section (Flags);
      return 0;
   end Marte_Getgroupclockid;

   ------------------
   --  Initialize  --
   ------------------
   procedure Initialize is
   begin
      pragma Debug (DBG.Assert (not Initialized));

      if Configuration_Parameters.Preallocated_Resources'First then
         for I in Free_Task_Sets'Range loop
            Task_Sets_Pool.Release (Free_Task_Sets (I)'Access);
         end loop;
         for I in Free_Task_Containers'Range loop
            Task_Containers_Pool.Release (Free_Task_Containers (I)'Access);
         end loop;
      end if;

      Initialized := True;
   end Initialize;

end Kernel.Task_Sets.Operations;
