------------------------------------------------------------------------------
-- --------------------        M a R T E   O S          ------------------- --
------------------------------------------------------------------------------
--                                                             V1.7     071213
--
--           'A d a . A p p l i c a t i o n _ S c h e d u l i n g'
--
--                                  Body
--
--
--  File 'a-appsch.adb'                                               By MAR.
--
--
--  Application-Defined Scheduling interface.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2007, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------

with System.OS_Interface; --  To use 'Self_Key'
with System.Task_Primitives.Operations;
with System.Tasking;
with Ada.Unchecked_Conversion;
with Ada.Real_Time.Delays; -- To use 'To_Duration (T : Real_Time.Time)'

with Kernel_Console;

with Basic_Integer_Types; use Basic_Integer_Types;
with Kernel.Tasks_Operations.Application_Scheduler;
with Kernel.Signals;
with Kernel.Timers;
with Hardware_Interface;

package body Ada.Application_Scheduling is

   package APPSCHD_DATA renames Kernel.Application_Scheduling_Data;
   package TO_APPSCHD   renames Kernel.Tasks_Operations.Application_Scheduler;
   package HWI          renames Hardware_Interface;

   use type Basic_Integer_Types.Int;

   ---------------------------------------------
   -- Conversion between 'Kernel.Task_Id' and --
   -- 'Ada.Task_Identification.Task_Id'       --
   ----------------------------------------------
   --
   function To_MaRTE_Task_Id (Ada_Task_Id : Ada.Task_Identification.Task_Id)
                             return Kernel.Task_Id;
   function To_MaRTE_Task_Id (Ada_Task_Id : Ada.Task_Identification.Task_Id)
                             return Kernel.Task_Id is

      function Pthread_T_To_MaRTE_Task_Id is new Ada.Unchecked_Conversion
        (System.OS_Interface.pthread_t, Kernel.Task_Id);
      function Convert_Ids is new Ada.Unchecked_Conversion
        (Ada.Task_Identification.Task_Id, System.Tasking.Task_Id);
   begin
      return
        Pthread_T_To_MaRTE_Task_Id
        (System.Task_Primitives.Operations.Get_Thread_Id
         (Convert_Ids (Ada_Task_Id)));
   end To_MaRTE_Task_Id;

   --------------------------------------------------
   --  Get_AppScheduler_Signo_Of_An_Appsched_Task  --
   --------------------------------------------------
   --
   --  'Kernel.Signals.Realtime_Signal'First + Appscheduler.Prio' as
   --  set in 'Ada.Application_Scheduling.Support.Install_Scheduler'
   function Get_AppScheduler_Signo_Of_An_Appsched_Task
     (T : Kernel.Task_Id) return Kernel.Signals.Signal;
   function Get_AppScheduler_Signo_Of_An_Appsched_Task (T : Kernel.Task_Id)
     return Kernel.Signals.Signal is
      function SA (Base : Kernel.Sched_App_Data_Base_Ac)
                  return APPSCHD_DATA.Sched_App_Data_Ac
        renames APPSCHD_DATA.UC_To_Sched_App_Data_Ac;
      use type Kernel.Signals.Signal;
   begin
      pragma Assert (Kernel.Tasks_Lists."/=" (SA (T.Sched_App).Scheduler,
                                              null));
      return Kernel.Signals.Realtime_Signal'First +
        Kernel.Signals.Signal (SA (T.Sched_App).Scheduler.Base_Prio);
   end Get_AppScheduler_Signo_Of_An_Appsched_Task;

   function To_Urgency (N : Natural) return Urgency is
   begin
      return Urgency (N);
   end To_Urgency;

   ----------------------------------------------------------------------------
   --  Scheduling Actions  ----------------------------------------------------
   ----------------------------------------------------------------------------

   procedure Raise_If_NZ (Ret : in Int);
   procedure Raise_If_NZ (Ret : in Int) is
   begin
      if Ret /= 0 then
         raise Scheduling_Actions_Full;
      end if;
   end Raise_If_NZ;

   ------------------
   --  Add_Accept  --
   ------------------
   procedure Add_Accept
     (Sched_Actions : in out Scheduling_Actions;
      Tid           : in Ada.Task_Identification.Task_Id) is
   begin
      Raise_If_NZ (TO_APPSCHD.POSIX_Appsched_Actions_Addaccept
                   (APPSCHD_DATA.Sched_Actions_Set
                    (Sched_Actions)'Unrestricted_Access,
                    To_MaRTE_Task_Id (Tid)));
   end Add_Accept;

   ------------------
   --  Add_Reject  --
   ------------------
   procedure Add_Reject
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id) is
   begin
      Raise_If_NZ (TO_APPSCHD.POSIX_Appsched_Actions_Addreject
                   (APPSCHD_DATA.Sched_Actions_Set
                    (Sched_Actions)'Unrestricted_Access,
                    To_MaRTE_Task_Id (Tid)));
   end Add_Reject;

   -----------------
   --  Add_Ready  --
   -----------------
   procedure Add_Ready
     (Sched_Actions : in out Scheduling_Actions;
      Tid           : in     Ada.Task_Identification.Task_Id) is
   begin
      Raise_If_NZ (TO_APPSCHD.POSIX_Appsched_Actions_Addactivate
                   (APPSCHD_DATA.Sched_Actions_Set
                    (Sched_Actions)'Unrestricted_Access,
                    To_MaRTE_Task_Id (Tid)));
   end Add_Ready;

   -----------------
   --  Add_Ready  --
   -----------------
   procedure Add_Ready
     (Sched_Actions : in out Scheduling_Actions;
      Tid           : in     Ada.Task_Identification.Task_Id;
      Urg           : in     Urgency) is
   begin
      Raise_If_NZ (TO_APPSCHD.POSIX_Appsched_Actions_Addactivateurg
                   (APPSCHD_DATA.Sched_Actions_Set
                    (Sched_Actions)'Unrestricted_Access,
                    To_MaRTE_Task_Id (Tid),
                    Kernel.Task_Urgency (Urg)));
   end Add_Ready;

   ---------------------------------
   --  Add_Timed_Task_Activation  --
   ---------------------------------
   procedure Add_Timed_Task_Activation
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id;
      Urg : in Urgency;
      At_Time : in Ada.Real_Time.Time) is
   begin
      Raise_If_NZ (TO_APPSCHD.POSIX_Appsched_Actions_Addtimedactivation
                   (APPSCHD_DATA.Sched_Actions_Set
                    (Sched_Actions)'Unrestricted_Access,
                    To_MaRTE_Task_Id (Tid),
                    Kernel.Task_Urgency (Urg),
                    HWI.Duration_To_HWTime
                    (Ada.Real_Time.Delays.To_Duration (At_Time))));
   end Add_Timed_Task_Activation;

   -------------------
   --  Add_Suspend  --
   -------------------
   procedure Add_Suspend
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id) is
   begin
      Raise_If_NZ (TO_APPSCHD.POSIX_Appsched_Actions_Addsuspend
                   (APPSCHD_DATA.Sched_Actions_Set
                    (Sched_Actions)'Unrestricted_Access,
                    To_MaRTE_Task_Id (Tid)));
   end Add_Suspend;

   -------------------
   --  Add_Timeout  --
   -------------------
   procedure Add_Timeout
     (Sched_Actions : in out Scheduling_Actions;
      At_Time : in Ada.Real_Time.Time) is
   begin
      pragma Unreferenced (Sched_Actions, At_Time);
      null;
      pragma Assert (False);
      --  Sched_Actions.Timeout_Active := True;
      --  Sched_Actions.Timeout := At_Time;
   end Add_Timeout;

   -----------------------------------
   --  Add_Timed_Task_Notification  --
   -----------------------------------
   procedure Add_Timed_Task_Notification
     (Sched_Actions : in out Scheduling_Actions;
      Tid           : in     Ada.Task_Identification.Task_Id;
      At_Time       : in     Ada.Real_Time.Time) is
   begin
      Raise_If_NZ (TO_APPSCHD.POSIX_Appsched_Actions_Addtimedtasknotification
                   (APPSCHD_DATA.Sched_Actions_Set
                    (Sched_Actions)'Unrestricted_Access,
                    To_MaRTE_Task_Id (Tid),
                    HWI.Duration_To_HWTime
                    (Ada.Real_Time.Delays.To_Duration (At_Time))));
   end Add_Timed_Task_Notification;

   ----------------------------
   --  Add_Timer_Expiration  --
   ----------------------------
   procedure Add_Timer_Expiration
     (Sched_Actions : in out Scheduling_Actions;
      T             : in out Ada.Real_Time.Execution_Time.Timer;
      Abs_Time      : in     Ada.Real_Time.Execution_Time.CPU_Time) is
      type Timer_Ac is access all Ada.Real_Time.Execution_Time.Timer;
      function To_Int is
         new Ada.Unchecked_Conversion (Timer_Ac, Basic_Integer_Types.Int);
   begin
      pragma Unreferenced (Sched_Actions);
      if not T.Is_Created then
         declare
            Event : Kernel.Signals.Signal_Event;
            Clock : aliased Kernel.Timers.Clock_Id;
            Ret : Basic_Integer_Types.Int;
         begin
            Ret :=
              Kernel.Timers.Pthread_Getcpuclockid (To_MaRTE_Task_Id (T.T.all),
                                                   Clock'Access);
            pragma Assert (Ret = 0);
            Event := (Event_Notification => Kernel.Signals.SIGNAL_NOTIFICATION,
                      Event_Signal =>
                        Get_AppScheduler_Signo_Of_An_Appsched_Task
                      (To_MaRTE_Task_Id (T.T.all)),
                      Event_Sigval => To_Int (T'Unrestricted_Access));
            T.Set_Low_Level_Timer (Kernel.Timers.Create_Timer (Clock, Event));
         end;
      end if;

      --  Arm absolute one-shot timer
      Kernel.Timers.Arm_Timer
        (T.Get_Low_Level_Timer, Kernel.Timers.ABSOLUTE_TIMER,
         Kernel.Timers.Timer_State'
         (Initial  => HWI.Duration_To_HWTime
          (Ada.Real_Time.Execution_Time.To_Duration (Abs_Time)),
          Interval => 0));
   end Add_Timer_Expiration;

   ----------------------------
   --  Add_Timer_Expiration  --
   ----------------------------
   procedure Add_Timer_Expiration
     (Sched_Actions : in out Scheduling_Actions;
      T             : in out Ada.Real_Time.Execution_Time.Timer;
      Interval      : in     Ada.Real_Time.Time_Span) is
      type Timer_Ac is access all Ada.Real_Time.Execution_Time.Timer;
      function To_Int is
         new Ada.Unchecked_Conversion (Timer_Ac, Basic_Integer_Types.Int);
   begin
      pragma Unreferenced (Sched_Actions);
      if not T.Is_Created then
         declare
            Event : Kernel.Signals.Signal_Event;
            Clock : aliased Kernel.Timers.Clock_Id;
            Ret : Basic_Integer_Types.Int;
         begin
            Ret :=
              Kernel.Timers.Pthread_Getcpuclockid (To_MaRTE_Task_Id (T.T.all),
                                                   Clock'Access);
            pragma Assert (Ret = 0);
            Event := (Event_Notification => Kernel.Signals.SIGNAL_NOTIFICATION,
                      Event_Signal =>
                        Get_AppScheduler_Signo_Of_An_Appsched_Task
                      (To_MaRTE_Task_Id (T.T.all)),
                      Event_Sigval => To_Int (T'Unrestricted_Access));
            T.Set_Low_Level_Timer (Kernel.Timers.Create_Timer (Clock, Event));
         end;
      end if;

      --  Arm relative one-shot timer
      Kernel.Timers.Arm_Timer
        (T.Get_Low_Level_Timer, Option_Sets.Empty_Option_Set,
         Kernel.Timers.Timer_State'
         (Initial  => HWI.Duration_To_HWTime
          (Ada.Real_Time.To_Duration (Interval)),
          Interval => 0));
   end Add_Timer_Expiration;

   ----------------------------------------------------------------------------
   --  Explicit Scheduler Invocation  -----------------------------------------
   ----------------------------------------------------------------------------

--   type Message_To_Scheduler_Ac is access Message_To_Scheduler'Class;
--   function To_Address is
--      new Ada.Unchecked_Conversion (Message_To_Scheduler_Ac, System.Address);

   --------------
   --  Invoke  --
   --------------
   procedure Invoke (Msg : access Message_To_Scheduler'Class) is
   begin
      if TO_APPSCHD.Invoke_Scheduler_With_Data (Msg.all'Address,
                                                Msg.all'Size / 8,
                                                Reply => null,
                                                Reply_Size => null) /= 0 then
         raise Task_Not_Application_Scheduled;
      end if;
   end Invoke;

   --------------
   --  Invoke  --
   --------------
   procedure Invoke (Msg   : access Message_To_Scheduler'Class;
                     Reply : access Reply_From_Scheduler'Class) is
      function To_Marte_Reply_Ac is
         new Ada.Unchecked_Conversion (System.Address,
                                       APPSCHD_DATA.Scheduler_Reply_Ac);
      Reply_Size : aliased APPSCHD_DATA.Scheduler_Reply_Size_T;
   begin
      if TO_APPSCHD.Invoke_Scheduler_With_Data
        (Msg.all'Address,
         Msg.all'Size / 8,
         To_Marte_Reply_Ac (Reply.all'Address),
         Reply_Size'Unchecked_Access) /= 0
      then
         raise Task_Not_Application_Scheduled;
      end if;
   end Invoke;

   ----------------------------------------------------------------------------
   --  Scheduler operations with null body  -----------------------------------
   ----------------------------------------------------------------------------

   procedure New_Task
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end New_Task;

   procedure Terminate_Task
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end Terminate_Task;

   procedure Ready
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end Ready;

   procedure Block
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end Block;

   procedure Yield
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end Yield;

   procedure Abort_Task
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end Abort_Task;

   procedure Change_Sched_Param
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end Change_Sched_Param;

   procedure Explicit_Call
     (Sched   : in out Scheduler;
      Tid     : in     Ada.Task_Identification.Task_Id;
      Msg     : access Message_To_Scheduler'Class;
      Reply   : access Reply_From_Scheduler'Class;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Msg, Reply, Actions);
      null;
   end Explicit_Call;

   procedure Task_Notification
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, Actions);
      null;
   end Task_Notification;

   procedure Timeout
     (Sched : in out Scheduler;
      Actions : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Actions);
      null;
   end Timeout;

   procedure Execution_Timer_Expiration
     (Sched         : in out Scheduler;
      Expired_Timer : in out Ada.Real_Time.Execution_Time.Timer;
      Actions       : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Expired_Timer, Actions);
      null;
   end Execution_Timer_Expiration;

   procedure Priority_Inherit
     (Sched          : in out Scheduler;
      Tid            : in     Ada.Task_Identification.Task_Id;
      From_Tid       : in     Ada.Task_Identification.Task_Id;
      Inherited_Prio : in     System.Any_Priority;
      Actions        : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, From_Tid, Inherited_Prio, Actions);
      null;
   end Priority_Inherit;

   procedure Priority_Uninherit
     (Sched            : in out Scheduler;
      Tid              : in     Ada.Task_Identification.Task_Id;
      From_Tid         : in     Ada.Task_Identification.Task_Id;
      Uninherited_Prio : in     System.Any_Priority;
      Actions          : in out Scheduling_Actions) is
   begin
      pragma Unreferenced (Sched, Tid, From_Tid, Uninherited_Prio, Actions);
      null;
   end Priority_Uninherit;

   ---------------------------------------------------------------------------
   --  Event masks  ----------------------------------------------------------
   ---------------------------------------------------------------------------
   function UC is new Ada.Unchecked_Conversion (Event_Code,
                                                APPSCHD_DATA.Event_Code_T);
   Ret : Int; --  Return value is not used so I prefer to do it global to
   --  improve footprint and speed
   procedure Fill (Mask : in out Event_Mask) is
   begin
      Ret := TO_APPSCHD.POSIX_Appsched_Fillset (Mask.Mask'Access);
      pragma Assert (Ret = 0);
   end Fill;

   procedure Empty (Mask : in out Event_Mask) is
   begin
      Ret := TO_APPSCHD.POSIX_Appsched_Fillset (Mask.Mask'Access);
      pragma Assert (Ret = 0);
      for E in Event_Code loop
         Ret :=
           TO_APPSCHD.POSIX_Appsched_Delset (Mask.Mask'Access, UC (E));
         pragma Assert (Ret = 0);
      end loop;
   end Empty;

   procedure Add (Event : in     Event_Code;
                  Mask  : in out Event_Mask) is
   begin
      Ret :=
        TO_APPSCHD.POSIX_Appsched_Addset (Mask.Mask'Access, UC (Event));
      pragma Assert (Ret = 0);
   end Add;

   procedure Delete (Event : in Event_Code;
                     Mask : in out Event_Mask) is
   begin
      if Event = Execution_Timer_Expiration or
        Event = Group_Timer_Expiration then
         Ret :=
           TO_APPSCHD.POSIX_Appsched_Delset (Mask.Mask'Access,
                                             APPSCHD_DATA.APPSCHED_SIGNAL);
      else
         Ret :=
           TO_APPSCHD.POSIX_Appsched_Delset (Mask.Mask'Access, UC (Event));
      end if;
      pragma Assert (Ret = 0);
   end Delete;

   function Is_Member (Event : in Event_Code;
                       Mask : in Event_Mask)
                      return Boolean is
   begin
      return TO_APPSCHD.POSIX_Appsched_Ismember (Mask.Mask, UC (Event)) = 0;
   end Is_Member;

   ----------------------
   --  Set_Event_Mask  --
   ----------------------
   procedure Set_Event_Mask (Mask : in Event_Mask) is
   begin
      Ret := TO_APPSCHD.POSIX_Appschedattr_Seteventmask (Mask.Mask);
   end Set_Event_Mask;

   ----------------------
   --  Get_Event_Mask  --
   ----------------------
   procedure Get_Event_Mask (Mask : out Event_Mask) is
   begin
      Ret := TO_APPSCHD.POSIX_Appschedattr_Geteventmask (Mask.Mask'Access);
   end Get_Event_Mask;

   --------------------------------
   --  Set_Protected_Event_Mask  --
   --------------------------------
   procedure Set_Protected_Event_Mask
     (Mask    : in Event_Mask;
      Ceiling : in System.Any_Priority) is
   begin
      pragma Unreferenced (Mask, Ceiling);
      null;

      Kernel_Console.Error ("Set_Protected_Event_Mask: no hace nada");
   end Set_Protected_Event_Mask;

   --------------------------------
   --  Get_Protected_Event_Mask  --
   --------------------------------
   procedure Get_Protected_Event_Mask
     (Mask    : out Event_Mask;
      Ceiling : out System.Any_Priority) is
   begin
      null;

      Mask := (Mask => Option_Sets.Empty_Option_Set);
      Ceiling := System.Any_Priority'First;
      Kernel_Console.Error ("Get_Protected_Event_Mask: no hace nada");
   end Get_Protected_Event_Mask;

   ----------------------------------------------------------------------------
   --  Change Task Scheduling Parameters  -------------------------------------
   ----------------------------------------------------------------------------

   function To_MaRTE_AppSched_Param_T_Ac is
         new Ada.Unchecked_Conversion (System.Address,
                                       Kernel.AppSched_Param_T_Ac);
   ----------------------
   --  Set_Parameters  --
   ----------------------
   procedure Set_Parameters
     (Tid   : in Ada.Task_Identification.Task_Id;
      Param : in Scheduling_Parameters'Class) is
      Ret : Int;
   begin
      Ret :=
       TO_APPSCHD.Pthread_Setappschedparam
        (To_MaRTE_Task_Id (Tid),
         To_MaRTE_AppSched_Param_T_Ac (Param'Address),
         Kernel.AppSched_Param_Size_T (Param'Size / 8));

      pragma Assert (Ret = 0);
   end Set_Parameters;

   ----------------------
   --  Get_Parameters  --
   ----------------------
   procedure Get_Parameters
     (Tid   : in     Ada.Task_Identification.Task_Id;
      Param : access Scheduling_Parameters'Class) is
      Ret : Int;
      AppSched_Param_Size : aliased Kernel.AppSched_Param_Size_T;
   begin
      Ret :=
        TO_APPSCHD.Pthread_Getappschedparam
        (To_MaRTE_Task_Id (Tid),
         To_MaRTE_AppSched_Param_T_Ac (Param.all'Address),
         AppSched_Param_Size'Access);

      pragma Assert (Ret = 0);
   end Get_Parameters;

end Ada.Application_Scheduling;
