------------------------------------------------------------------------------
--                             IDE body                                     --
------------------------------------------------------------------------------
--  The implementation is based on the ATA LBA28 standard.                  --
--  The main document I have used to make the driver is the CompactFlash    --
--  especification (http://www.compactflash.org) (note that the driver is   --
--  also compatible with CF cards in True Ide Mode).                        --
--  The module has been tested on QEMU emulator and a CompactFlash on a     --
--  PC-104 mother board so far.                                             --
--  How to use it: test_ide.adb                                             --
--                                            daniel.sangorrin_at_gmail.com --
------------------------------------------------------------------------------
--  for System.Address
with System;
--  for Insw, Outsw
with Io_Interface; use Io_Interface;
pragma Elaborate_All (Io_Interface);
--  for the Interrupt handler:
with Marte_Hardware_Interrupts;
pragma Elaborate_All (Marte_Hardware_Interrupts);
--  for synchronizing IRQ_handler with the task
with Marte_Semaphores;
pragma Elaborate_All (Marte_Semaphores);
--  for timing funcions
with Ada.Real_Time; use Ada.Real_Time;
with Type_Timespec; use Type_Timespec;
with Kernel.Tasks_Operations.Nanosleep_Gnat;
--  for output messages
with Basic_console_io; use Basic_console_io;

package body IDE is

   package Mhi renames Marte_Hardware_Interrupts;
   package TON renames Kernel.Tasks_Operations.Nanosleep_Gnat;

   use type BIT.Unsigned_8;
   use type BIT.Unsigned_32;

   D400NS : constant := 0.0000004; --  400ns of delay (see ATA standard)

   ----------------------
   -- IO ADDRESS SPACE --
   ----------------------

   BASE : constant array (Drive) of Io_Port := (16#1F0#, 16#1F0#, 16#170#, 16#170#);

   IDE_DATA         : constant := 0; --  RW - 16 bit - data blocks
   IDE_ERROR        : constant := 1; --  RO - BBK/CRC.UNC.0.IDNF.0.ABRT.0.AMNF
   IDE_SECTOR_COUNT : constant := 2; --  RW - 0=256 - return success rw
   IDE_SECTOR_START : constant := 3; --  RW - LBA[7:0]
   IDE_LOW_CYL      : constant := 4; --  RW - LBA[15:8]
   IDE_HIGH_CYL     : constant := 5; --  RW - LBA[23:16]
   IDE_DRIVE_HEAD   : constant := 6; --  RW - 1.LBA.1.DRIVE.LBA[27:24]
   IDE_STATUS       : constant := 7; --  RO - BUSY.RDY.DWF.DSC.DRQ.CORR.0.ERR - Clear IRQ
   IDE_FEATURES     : constant := 1; --  WO
   IDE_COMMAND      : constant := 7; --  WO - See commands
   IDE_DEVICE_CTRL  : constant := 16#206#; --  WO - 0.0.0.0.0.SOFTRESET.IRQ.0
   IDE_STATUS_AUX   : constant := 16#206#; -- RO - Like Status - Doesn't clear IRQ

   ----------------
   -- Interrupts --
   ----------------

   IRQs : constant array (Drive) of Mhi.Hardware_Interrupt :=
     (hda => Mhi.FIXED_DISK_INTERRUPT, hdb => Mhi.FIXED_DISK_INTERRUPT, --  IRQ = 14
      hdc => Mhi.RESERVED4_INTERRUPT, hdd => Mhi.RESERVED4_INTERRUPT);  --  IRQ = 15

   Sem  : aliased Marte_Semaphores.Semaphore;

   function IDE_IRQ_Handler (
         Area : in     System.Address;
         Intr : in     Mhi.Hardware_Interrupt)
     return Mhi.Handler_Return_Code is
     Value : aliased Marte_Semaphores.Semaphore_Value;
     use type Marte_Semaphores.Semaphore_Value;
     Ret : BIT.Int;
   begin
      --  Put ("Debug: Intr called"); New_Line;
      Ret := Marte_Semaphores.Getvalue (Sem'access, Value'access);
      if Value = 0 then
         Ret := Marte_Semaphores.Post (Sem'access);
      end if;
      return Mhi.Posix_Intr_Handled_Do_Not_Notify;
   end IDE_IRQ_Handler;

   ----------------
   -- delay400NS --
   ----------------

   procedure delay400NS is
      Interval, Remaining : aliased Duration;
      Ret : BIT.Int;
   begin
      Interval := D400NS;
      Ret := TON.Nanosleep_Gnat (Interval'Unchecked_Access, Remaining'Unchecked_Access);
   end delay400NS;

   -------------------------
   -- Is_Controller_Ready --
   -------------------------

   function Is_Controller_Ready (Drv : in Drive) return BIT.Int is
      Byte : BIT.Unsigned_8;
      Mx_Retry : constant := 100000;
   begin
      for I in 1..Mx_Retry loop
         Byte := Inb_P (BASE(Drv) + IDE_STATUS);
         if (Byte and 2#1101_0000#) = 2#0101_0000# then
            --  Put ("Debug: Retrys in Is_Controller_Ready = "&Integer'Image(I-1)); New_Line;
            return OK;
         end if;
      end loop;
      return ERROR;
   end Is_Controller_Ready;

   -------------------------
   -- Check_Data_Transfer --
   -------------------------

   function Check_Data_Transfer (Drv : in Drive) return BIT.Int is
      Byte : BIT.Unsigned_8;
      Mx_Retry : constant := 100000;
   begin
      for I in 1..Mx_Retry loop
         Byte := Inb_P (BASE(Drv) + IDE_STATUS);
         if (Byte and 2#0001_0001#) = 2#0001_0001# then
            return ERROR;
         elsif (Byte and 2#1101_1001#) = 2#0101_1000# then
            --  Put ("Debug: Retrys in Check_Data_Transfer = "&Integer'Image(I-1)); New_Line;
            return OK;
         end if;
      end loop;
      return ERROR;
   end Check_Data_Transfer;

   ----------
   -- Init --
   ----------

   function Init (Drv  : in  Drive) return BIT.Int is
   begin
      if Marte_Semaphores.Initialize (Sem'access,0,0) /= 0 then
         Put("Error: Could not initialize the semaphore"); New_Line;
         return ERROR;
      end if;
      if Mhi.Associate (IRQs (Drv), IDE_IRQ_Handler'access, System.Null_Address, 0) /= 0 then
         Put("Error: Could not associate to the IRQ"); New_Line;
         return ERROR;
      end if;
      if Mhi.Unlock (IRQs (Drv)) /= 0 then
         Put("Error: Could not Lock the IRQ"); New_Line;
         return ERROR;
      end if;
      --  Select the Drive and LBA mode
      case Drv is
         when hda | hdc =>
            Outb_P (BASE (Drv) + IDE_DRIVE_HEAD, 2#1110_0000#);
         when hdb | hdd =>
            Outb_P (BASE (Drv) + IDE_DRIVE_HEAD, 2#1111_0000#);
      end case;
      --  Interrupts enabled
      Outb_P (BASE(Drv) + IDE_DEVICE_CTRL, 2#0000_0000#);
      delay400NS;
      --  Wait for the IDE controller
      return Is_Controller_Ready (Drv);
   end Init;

   -----------
   -- Reset --
   -----------

   function Reset (Drv : in Drive) return BIT.Int is
   begin
      --  Select the Drive
      case Drv is
         when hda | hdc =>
            Outb_P (BASE (Drv) + IDE_DRIVE_HEAD, 2#1110_0000#);
         when hdb | hdd =>
            Outb_P (BASE (Drv) + IDE_DRIVE_HEAD, 2#1111_0000#);
      end case;
      --  Soft Reset + Interrupts enabled
      Outb_P (BASE (Drv) + IDE_DEVICE_CTRL, 2#0000_0100#);
      delay400NS;
      Outb_P (BASE (Drv) + IDE_DEVICE_CTRL, 2#0000_0000#);
      delay400NS;
      --  Wait for the IDE controller
      return Is_Controller_Ready (Drv);
   end Reset;

   ----------------
   -- RW_Sectors --
   ----------------

   function RW_Sectors
     (Drv    : in     Drive;
      RW     : in     Read_Write;
      Start  : in     Sector_Offset; --  physical offset
      Count  : in     Sector_Count;  --  0=256 sectors
      Buffer : in     FSDT.Buffer_Ac)
      return BIT.Int
   is
      Sec, Cyl_L, Cyl_H, Head : BIT.Unsigned_8;
      Pointer : System.Address;
      Words : BIT.Unsigned_32;
   begin
      if Is_Controller_Ready (Drv) /= OK then
         return ERROR;
      end if;
      --  Parse the LBA address
      Sec   := BIT.Unsigned_8 (Start and 16#FF#);
      Cyl_L := BIT.Unsigned_8 (BIT.Shift_Right (Start, 8) and 16#FF#);
      Cyl_H := BIT.Unsigned_8 (BIT.Shift_Right (Start, 16) and 16#FF#);
      Head  := BIT.Unsigned_8 (BIT.Shift_Right (Start, 24) and 16#F#);
      --  Select the Drive and Set LBA[27:24]
      case Drv is
         when hda | hdc =>
            Outb_P (BASE(Drv) + IDE_DRIVE_HEAD, 2#1110_0000# or Head);
         when hdb | hdd =>
            Outb_P (BASE(Drv) + IDE_DRIVE_HEAD, 2#1111_0000# or Head);
      end case;
      --  Set how many sectors we want to read or write
      Outb_P (BASE(Drv) + IDE_SECTOR_COUNT, Count);
      --  Set LBA[0:23]
      Outb_P (BASE(Drv) + IDE_SECTOR_START, Sec );
      Outb_P (BASE(Drv) + IDE_LOW_CYL, Cyl_L);
      Outb_P (BASE(Drv) + IDE_HIGH_CYL, Cyl_H);
      --  Use Set_Multiple command to enable Multiple sector transfer
      Outb_P (BASE(Drv) + IDE_COMMAND, 16#C6#); delay400NS;
      --  Send the Command (I use always Multiple commands)
      if RW = Read then
         Outb_P (BASE(Drv) + IDE_COMMAND, 16#C4#); delay400NS;
      else --  Write
         Outb_P (BASE(Drv) + IDE_COMMAND, 16#C5#); delay400NS;
      end if;
      if Marte_Semaphores.Timedwait (Sem'access, Clock + To_Time_Span (5.0)) /= 0 then
         Put("Error: waiting the semaphore"); New_Line;
         return ERROR;
      end if;
      --  Check the operation for errors
      if Check_Data_Transfer (Drv) /= OK then
         return ERROR;
      end if;
      --  Read/Write the words
      Pointer := Buffer (1)'Address;
      if Count = 0 then
         Words := 256*SECTOR_SIZE/2; -- 0=256 sectors,  a_sector=256 words
      else
         Words := BIT.Unsigned_32 (Count)*SECTOR_SIZE/2;
      end if;
      if RW = Read then
         Insw (BASE(Drv) + IDE_DATA, Pointer, Words);
      else
         Outsw (BASE(Drv) + IDE_DATA, Pointer, Words);
      end if;
      return OK;
   end RW_Sectors;

   ----------
   -- Free --
   ----------

   function Free (Drv  : in Drive) return BIT.Int is
   begin
      --  Select the Drive
      case Drv is
         when hda | hdc =>
            Outb_P (BASE (Drv) + IDE_DRIVE_HEAD, 2#1110_0000#);
         when hdb | hdd =>
            Outb_P (BASE (Drv) + IDE_DRIVE_HEAD, 2#1111_0000#);
      end case;
      --  Disable Interrupts
      Outb_P (BASE(Drv) + IDE_DEVICE_CTRL, 2#0000_0010#);
      delay400NS;
      if Mhi.Lock (IRQs (Drv)) /= 0 then
         Put("Error: Could not Lock the IRQ"); New_Line;
         return ERROR;
      end if;
      if Mhi.Disassociate (IRQs (Drv), IDE_IRQ_Handler'access) /= 0 then
         Put("Error: Could not disassociate to the IRQ"); New_Line;
         return ERROR;
      end if;
      if Marte_Semaphores.Destroy (Sem'access) /= 0 then
         Put("Error: Could not destroy the semaphore"); New_Line;
         return ERROR;
      end if;
      return OK;
   end Free;

end IDE;