--  ------------------------------------------------------------------------
--  ---------------------           RT-EP           ------------------------
--  ------------------------------------------------------------------------
--                         Copyright (C) 2003-2005
--                     Universidad de Cantabria, SPAIN
--                        http://www.ctr.unican.es/
--
--    This program is free software; you can redistribute it and/or
--    modify it under the terms of the GNU General Public
--    License as published by the Free Software Foundation; either
--    version 2 of the License, or (at your option) any later version.
--
--    This program is distributed in the hope that it will be useful,
--    but WITHOUT ANY WARRANTY; without even the implied warranty of
--    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
--    General Public License for more details.
--
--    You should have received a copy of the GNU General Public
--    License along with this program; if not, write to the
--    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
--    Boston, MA 02111-1307, USA.
--
--    As a special exception, if other files instantiate generics from
--    this unit, or you link this unit with other files to produce an
--    executable, this  unit  does not  by itself cause  the resulting
--    executable to be covered by the GNU General Public License. This
--    exception does not however invalidate any other reasons why the
--    executable file  might be covered by the  GNU Public License.
--  ------------------------------------------------------------------------
--                                                     RT-EP /Ada95 {V1.1}
--
--             'r t e p - p r o t o c o l - c h a n n e l s . a d b'
--
--                                     Ada
--
--
--  File 'rtep-protocol-channels.adb'                              By Chema.
--                                                       Jose Maria Martinez
--                                                       <chema@gmx.net>
--
--  RT-EP Channel handling. These functions will be used to store info in a
--  channel
--
--
--
--
-----------------------------------------------------------------------------

with RTEP.Protocol.Packet;
with RTEP.Protocol.Servers;
with Priority_FIFO_Queues; pragma Elaborate_All (Priority_FIFO_Queues);
with Queues; pragma Elaborate_All (Queues);

package body RTEP.Protocol.Buffers is

   -----------------------
   -- 0) Initialization --
   -----------------------
   procedure Buffers_Init is
   begin
      for Id in Server_ID'Range loop
         Server_Queues_Init (Id);
      end loop;
   end Buffers_Init;

   -----------------------
   -- 1) CHANNEL QUEUES --
   -----------------------
   --  There is one transmission channel plus several reception channels (we
   --  have also the queues for the servers. We use a generic ADT (abstract
   --  data type) that implemments priority queues with a FIFO behaviour.
   --  Access to the queues is protected through the use of a monitor.
   subtype Index is Integer range 0 .. RTEP.Number_Of_Channels;
   Tx_Channel : constant Index := 0; --  Channel 0 is reserved for transmission

   type FIFO_Index is mod RTEP.Max_Queued_Element_Same_Priority;
   package Prio_FIFO_Queues_Pkg is new Priority_FIFO_Queues
     (FIFO_Index => FIFO_Index,
      Element  => Packet.RTEP_Packet_Ac,
      Priority => RTEP.Priority,
      ">"      => RTEP.">");

   type RTEP_Queues is array (Index) of Prio_FIFO_Queues_Pkg.Prio_Queue;

   protected Queues_Monitor is

      pragma Priority (RTEP.RTEP_Task_Prio);

      procedure Queue_Monitor_Insert
        (RTEP_Channel : in Index;
         Packet_Ac : in Packet.RTEP_Packet_Ac;
         Prio : in Priority);

      entry Queue_Monitor_Extract
        (RTEP_Channel : in Index;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Prio : out Priority;
         Blocking : in Boolean);

      procedure Queue_Monitor_Read_First
        (RTEP_Channel : in Index;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Prio : out Priority);
   private
      entry Queue_Monitor_Blocking_Extract (Index)
        (RTEP_Channel : in Index;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Prio : out Priority;
         Blocking : in Boolean);
      The_RTEP_Queues : RTEP_Queues;
   end Queues_Monitor;

   protected body Queues_Monitor is

      procedure Queue_Monitor_Insert
        (RTEP_Channel : in Index;
         Packet_Ac : in Packet.RTEP_Packet_Ac;
         Prio : in Priority) is
      begin
         Prio_FIFO_Queues_Pkg.Enqueue
            (The_RTEP_Queues (RTEP_Channel), Packet_Ac, Prio);
      end Queue_Monitor_Insert;

      entry Queue_Monitor_Extract
        (RTEP_Channel : in Index;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Prio : out Priority;
         Blocking : in Boolean) when True is
      begin
         if Prio_FIFO_Queues_Pkg.Empty (The_RTEP_Queues (RTEP_Channel)) then
            --  If the queue is Empty we have to wait only if Blocking = True.
            if  Blocking then
               requeue Queue_Monitor_Blocking_Extract (RTEP_Channel);
            else
               Packet_Ac := null;
            end if;
         else
            Prio_FIFO_Queues_Pkg.Dequeue
               (The_RTEP_Queues (RTEP_Channel), Packet_Ac, Prio);
         end if;
      end Queue_Monitor_Extract;

      entry Queue_Monitor_Blocking_Extract (for I in Index)
        (RTEP_Channel : in Index;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Prio : out Priority;
         Blocking : in Boolean)
      when not Prio_FIFO_Queues_Pkg.Empty (The_RTEP_Queues (I)) is
      begin
         if Blocking then
            Prio_FIFO_Queues_Pkg.Dequeue
               (The_RTEP_Queues (RTEP_Channel), Packet_Ac, Prio);
         end if;
      end Queue_Monitor_Blocking_Extract;

      procedure Queue_Monitor_Read_First
        (RTEP_Channel : in Index;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Prio : out Priority) is
      begin
         if Prio_FIFO_Queues_Pkg.Empty (The_RTEP_Queues (RTEP_Channel)) then
            Packet_Ac := Null;
         else
            Prio_FIFO_Queues_Pkg.Read_First
               (The_RTEP_Queues (RTEP_Channel), Packet_Ac, Prio);
         end if;
      end Queue_Monitor_Read_First;
   end Queues_Monitor;

   ------------------
   -- 2) Rx QUEUES --
   ------------------
   --  Rx_Queues_Insert_Packet
   procedure Rx_Queues_Insert_Packet
     (RTEP_Channel   : in Channel;
      Packet_Ac : in Packet.RTEP_Packet_Ac;
      Prio           : in Priority) is
   begin
      Queues_Monitor.Queue_Monitor_Insert
         (Index (RTEP_Channel), Packet_Ac, Prio);
   exception
      when Prio_FIFO_Queues_Pkg.Priority_Storage_Full |
           Prio_FIFO_Queues_Pkg.Full_Queue =>
         raise Full_Of_Packets;
   end Rx_Queues_Insert_Packet;

   --  Rx_Queues_Extract_Packet
   procedure Rx_Queues_Extract_Packet
     (RTEP_Channel   : in Channel;
      Packet_Ac : out Packet.RTEP_Packet_Ac;
      Prio           : out Priority;
      Blocking       : in Boolean := True) is
   begin
      Queues_Monitor.Queue_Monitor_Extract
         (Index (RTEP_Channel), Packet_Ac, Prio, Blocking);
   exception
      when Prio_FIFO_Queues_Pkg.Empty_Queue =>
         raise Empty;
   end Rx_Queues_Extract_Packet;

   --  Rx_Queues_Read_Packet
   procedure Rx_Queues_Read_Packet
     (RTEP_Channel   : in Channel;
      Packet_Ac : out Packet.RTEP_Packet_Ac;
      Prio           : out Priority) is
   begin
      Queues_Monitor.Queue_Monitor_Read_First
         (Index (RTEP_Channel), Packet_Ac, Prio);
   end Rx_Queues_Read_Packet;

   -------------------------
   -- 3) General Tx QUEUE --
   -------------------------
   --  Tx_Queue_Insert_Packet
   procedure Tx_Queue_Insert_Packet
     (Packet_Ac : in Packet.RTEP_Packet_Ac;
      Prio           : in Priority) is
   begin
      Queues_Monitor.Queue_Monitor_Insert
         (Tx_Channel, Packet_Ac, Prio);
   exception
      when Prio_FIFO_Queues_Pkg.Priority_Storage_Full |
           Prio_FIFO_Queues_Pkg.Full_Queue =>
         raise Full_Of_Packets;
   end Tx_Queue_Insert_Packet;

   --  Tx_Queue_Extract_Packet
   procedure Tx_Queue_Extract_Packet
     (Packet_Ac : out Packet.RTEP_Packet_Ac;
      Prio           : out Priority;
      Blocking       : in Boolean := True) is
   begin
      Queues_Monitor.Queue_Monitor_Extract
         (Tx_Channel, Packet_Ac, Prio, Blocking);
   exception
      when Prio_FIFO_Queues_Pkg.Empty_Queue =>
         raise Empty;
   end Tx_Queue_Extract_Packet;

   --  Tx_Queue_Read_Packet
   procedure Tx_Queue_Read_Packet
     (Packet_Ac : out Packet.RTEP_Packet_Ac;
      Prio           : out Priority) is
   begin
      Queues_Monitor.Queue_Monitor_Read_First
         (Tx_Channel, Packet_Ac, Prio);
   end Tx_Queue_Read_Packet;

   -------------------------
   -- 4) Server Tx QUEUES --
   -------------------------
   --  There is a normal FIFO Queue for each Server.
   --  We use a monitor to protect the access to these Queues.

   package Tx_Queues_Pkg is new Queues
     (Max_Queued_Element_Same_Priority, Packet.RTEP_Packet_Ac);

   type Server_Tx_Queues_Type is array (Server_ID) of Tx_Queues_Pkg.Queue;

   protected Tx_Queues_Monitor is
      pragma Priority (RTEP.RTEP_Task_Prio);
      procedure Tx_Queues_Monitor_Insert
        (Id             : in Server_Id;
         Packet_Ac : in Packet.RTEP_Packet_Ac);
      entry Tx_Queues_Monitor_Extract
        (Id             : in Server_Id;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Blocking       : in Boolean := True);
      procedure Tx_Queues_Monitor_Init
         (Id             : in Server_Id);
   private
      entry Tx_Queues_Monitor_Blocking_Extract (Server_Id)
        (Id             : in Server_Id;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Blocking       : in Boolean := True);

      Server_Tx_Queues : Server_Tx_Queues_Type;
   end Tx_Queues_Monitor;

   protected body Tx_Queues_Monitor is
      procedure Tx_Queues_Monitor_Insert
        (Id             : in Server_ID;
         Packet_Ac : in Packet.RTEP_Packet_Ac) is
      begin
         Tx_Queues_Pkg.Enqueue (Packet_Ac,
                                Server_Tx_Queues (Id));
      end Tx_Queues_Monitor_Insert;

      entry Tx_Queues_Monitor_Extract
        (Id            : in Server_Id;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Blocking : in Boolean := True) when True is
      begin
         if Tx_Queues_Pkg.Empty (Server_Tx_Queues(Id)) then
            --  If the queue is Empty we have to wait only if Blocking = True.
            if Blocking then
               requeue Tx_Queues_Monitor_Blocking_Extract (Id);
            else
               Packet_Ac := null;
            end if;
         else
            Tx_Queues_Pkg.Dequeue (Packet_Ac, Server_Tx_Queues(Id));
         end if;
      end Tx_Queues_Monitor_Extract;

      entry Tx_Queues_Monitor_Blocking_Extract (for I in Server_Id)
        (Id             : in Server_Id;
         Packet_Ac : out Packet.RTEP_Packet_Ac;
         Blocking       : in Boolean := True)
         when not Tx_Queues_Pkg.Empty(Server_Tx_Queues(I)) is
      begin
         if Blocking then
            Tx_Queues_Pkg.Dequeue (Packet_Ac, Server_Tx_Queues(Id));
         end if;
      end Tx_Queues_Monitor_Blocking_Extract;

      procedure Tx_Queues_Monitor_Init
         (Id             : in Server_Id) is
      begin
         Tx_Queues_Pkg.Init (Server_Tx_Queues(Id));
      end Tx_Queues_Monitor_Init;
   end Tx_Queues_Monitor;

   --------------------------
   -- Server_Insert_Packet --
   --------------------------

   procedure Server_Queues_Insert_Packet
     (Packet_Ac : in Packet.RTEP_Packet_Ac;
      Id             : in Server_Id) is
   begin
      Tx_Queues_Monitor.Tx_Queues_Monitor_Insert (Id, Packet_Ac);
      RTEP.Protocol.Servers.Packet_Enqueued (Id);
   exception
      when Tx_Queues_Pkg.Full_Queue =>
         raise Full_Of_Packets;
   end Server_Queues_Insert_Packet;

   ---------------------------
   -- Server_Extract_Packet --
   ---------------------------

   procedure Server_Queues_Extract_Packet
     (Packet_Ac : out Packet.RTEP_Packet_Ac;
      Id             : in Server_Id;
      Blocking       : in Boolean := True) is
   begin
      Tx_Queues_Monitor.Tx_Queues_Monitor_Extract
         (Id, Packet_Ac, Blocking);
      RTEP.Protocol.Servers.Packet_Sent
         (Id, Packet.Get_Timestamp (Packet_Ac.all));
   exception
      when  Tx_Queues_Pkg.Empty_Queue =>
         raise Empty;
   end Server_Queues_Extract_Packet;

   ------------------------
   -- Server_Queues_Init --
   ------------------------

   procedure Server_Queues_Init
     (Id  : in Server_Id) is
   begin
      Tx_Queues_Monitor.Tx_Queues_Monitor_Init (Id);
   end Server_Queues_Init;

end RTEP.Protocol.Buffers;
